
from setuptools import setup, find_packages


setup(
    name="pyDGEA",
    version='0.0.2',
    python_requires=">=3.9.0",
    license="MIT",
    description="A python package that implements differential gene expression analysis.",
    author="Innocent Labs",
    author_email="jakec63738@gmail.com",
    packages=find_packages()
)